import numpy as np
import sys
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys,gene_to_guides_fast

def s_loop(self):
    if self.verbose:
        print("Updating s values.",file=sys.stdout)
    for s_triple in self.hash_s.keys():
        indices=collect_all_keys(self,s_triple)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: s_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: s_denominator(self,x),indices))))
        new_s=((self.prio_tau_s*self.prio_mu_s)+numerator_sum)/(self.prio_tau_s+denominator_sum)
        if self.debug: print(f"Old s {self.s[self.hash_s[s_triple]]}.", sys.stdout)
        self.s[self.hash_s[s_triple]] = new_s
        if self.debug: print(f"New s {self.s[self.hash_s[s_triple]]}.", sys.stdout)

def ss_loop(self):
    if self.verbose:
        print("Updating ss values.",file=sys.stdout)
    for s_triple in self.hash_s.keys():
        indices=collect_all_keys(self,s_triple)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: s_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: s_denominator(self,x),indices))))
        new_ss=((self.prio_tau_ss*self.prio_mu_ss)+numerator_sum)/(self.prio_tau_ss+denominator_sum)
        if self.debug: print(f"Old ss {self.ss[self.hash_s[s_triple]]}.", sys.stdout)
        self.ss[self.hash_s[s_triple]] = new_ss
        if self.debug: print(f"New ss {self.ss[self.hash_s[s_triple]]}.", sys.stdout)

def s_numerator(self,key_dict):
    numerator=(self.tau[self.hash_t[key_dict["t"]]]*self.r[self.hash_r[key_dict["r"]]])*\
    (self.D[self.hash_d[key_dict["t"]]]-\
    (self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]])-\
    (self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]])-\
    (self.x[self.hash_x[key_dict["gi"]]]*self.y[self.hash_y[key_dict["gl"]]]))
    return numerator

def s_denominator(self,key_dict):
    denominator = self.tau[self.hash_t[key_dict["t"]]] * self.rr[self.hash_r[key_dict["r"]]]
    return denominator

def collect_all_keys(self,s_triple):
    #only get keys from tau and split them accordingly
    gi, hj, fk, l = s_triple.split("_")
    s_triple="_".join([gi, hj, fk])
    s_as_guides=[gene_to_guides_fast(self,x) for x in s_triple.split("_")]
    s_as_guides = [x for y in s_as_guides for x in y]
    tau_keys = []
    for k in self.hash_t.keys():
        gi, hj, fk, l = k.split("_")
        if gi in s_as_guides and hj in s_as_guides and fk in s_as_guides:
            tau_keys.append(k)
        else:
            continue
    out=[]
    for k in tau_keys:
        gi,hj,fk,l = k.split("_")
        gl = self.guide_to_gene(gi)
        hl = self.guide_to_gene(hj)
        fl = self.guide_to_gene(fk)
        s = find_s_keys(self,gl,hl,fl,l)
        r = find_r_keys(self,gi,hj,fk)
        out.append({"t":k,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s})
    return out


import sys
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys

def tau_loop(self):
    if self.verbose:
        print("Updating tau values.",file=sys.stdout)
    for tau_triple in self.hash_t.keys():
        indices=collect_all_keys(self,tau_triple)
        new_alpha = tau_alpha(self,indices)
        self.alpha[self.hash_t[tau_triple]] = new_alpha
        new_beta=tau_beta(self,indices)
        self.beta[self.hash_t[tau_triple]] = new_beta
        new_tau=new_alpha/new_beta
        if self.debug: print(f"Old tau {self.tau[self.hash_t[tau_triple]]}.", sys.stdout)
        self.tau[self.hash_t[tau_triple]] = new_tau
        if self.debug: print(f"New s {self.tau[self.hash_t[tau_triple]]}.", sys.stdout)

def tau_alpha(self,key_dict):
    return self.alpha[self.hash_t[key_dict["t"]]]+0.5

def tau_beta(self,key_dict):
    beta=self.beta[self.hash_t[key_dict["t"]]]
    D2=0.5 * self.D[self.hash_d[key_dict["t"]]]**2
    Dxgiygl=self.D[self.hash_d[key_dict["t"]]]*self.x[self.hash_x[key_dict["gi"]]]*self.y[self.hash_y[key_dict["gl"]]]
    Dxhjyhl=self.D[self.hash_d[key_dict["t"]]]*self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]]
    Dxfkyfl=self.D[self.hash_d[key_dict["t"]]]*self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]]
    Drs=self.D[self.hash_d[key_dict["t"]]]*self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]
    xgi2ygl2=self.xx[self.hash_x[key_dict["gi"]]]*self.yy[self.hash_y[key_dict["gl"]]]
    xgiyglxhjyhl=self.x[self.hash_x[key_dict["gi"]]]*self.y[self.hash_y[key_dict["gl"]]]*self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]]
    xgiyglxfkyfl=self.x[self.hash_x[key_dict["gi"]]]*self.y[self.hash_y[key_dict["gl"]]]*self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]]
    xgiyglrs=2*self.x[self.hash_x[key_dict["gi"]]]*self.y[self.hash_y[key_dict["gl"]]]*self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]
    xhj2yhl2=0.5*self.xx[self.hash_x[key_dict["hj"]]]*self.yy[self.hash_y[key_dict["hl"]]]
    xhjyhlxfkyfl=self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]]*self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]]
    xhjyhlrs=self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]]*self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]
    r2s2=0.5*self.rr[self.hash_r[key_dict["r"]]]*self.ss[self.hash_s[key_dict["s"]]]
    out=beta+D2-Dxgiygl-Dxhjyhl-Dxfkyfl-Drs+xgi2ygl2+xgiyglxhjyhl+xgiyglxfkyfl+xgiyglrs+xhj2yhl2+xhjyhlxfkyfl+xhjyhlrs+r2s2
    return out


def collect_all_keys(self,tau_triple):
    gi,hj,fk,l = tau_triple.split("_")
    gl = self.guide_to_gene(gi)
    hl = self.guide_to_gene(hj)
    fl = self.guide_to_gene(fk)
    s = find_s_keys(self,gl,hl,fl,l)
    r = find_r_keys(self,gi,hj,fk)
    out={"t":tau_triple,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s}
    return out


import numpy as np
import sys
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys,gene_to_guides_fast

def y_loop(self):
    if self.verbose:
        print("Updating y values.",file=sys.stdout)
    for gl in self.hash_y.keys():
        indices=collect_all_keys(self,gl)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: y_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: y_denominator(self,x),indices))))
        new_ygl=((self.prio_tau_y*self.prio_mu_y)+numerator_sum)/(self.prio_tau_y+denominator_sum)
        if self.debug: print(f"Old ygl {self.y[self.hash_y[gl]]}.", sys.stdout)
        self.y[self.hash_y[gl]] = new_ygl
        if self.debug: print(f"New ygl {self.y[self.hash_y[gl]]}.", sys.stdout)


def yy_loop(self):
    if self.verbose:
        print("Updating yy values.",file=sys.stdout)
    for gl in self.hash_y.keys():
        indices=collect_all_keys(self,gl)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: y_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: y_denominator(self,x),indices))))
        new_yygl=((self.prio_tau_yy*self.prio_mu_yy)+numerator_sum)/(self.prio_tau_yy+denominator_sum)
        if self.debug: print(f"Old yygl {self.yy[self.hash_y[gl]]}.", sys.stdout)
        self.yy[self.hash_y[gl]] = new_yygl
        if self.debug: print(f"New yygl {self.yy[self.hash_y[gl]]}.", sys.stdout)

def y_numerator(self,key_dict):
    numerator=(self.tau[self.hash_t[key_dict["t"]]]*self.x[self.hash_x[key_dict["gi"]]])*\
    (self.D[self.hash_d[key_dict["t"]]]-\
    (self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]])-\
    (self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]])-\
    (self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]))
    return numerator

def y_denominator(self,key_dict):
    denominator = self.tau[self.hash_t[key_dict["t"]]] * self.xx[self.hash_x[key_dict["gi"]]]
    return denominator

def collect_all_keys(self,gl_extern):
    #only get keys from tau and split them accordingly
    gene,sample=gl_extern.split("_")
    gi_extern=gene_to_guides_fast(self,gene)
    tau_keys = [x for x in self.hash_t.keys() if any(list(map(lambda d: d in x,gi_extern)))]
    out=[]
    for k in tau_keys:
        gi,hj,fk,l = k.split("_")
        gl = self.guide_to_gene(gi)
        hl = self.guide_to_gene(hj)
        fl = self.guide_to_gene(fk)
        s = find_s_keys(self,gl,hl,fl,l)
        r = find_r_keys(self,gi,hj,fk)
        out.append({"t":k,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s})
    return out


import numpy as np
import sys
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys

def x_loop(self):
    if self.verbose:
        print("Updating x values",file=sys.stdout)
    for gi in self.unique_guides:
        indices=collect_all_keys(self,gi)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: x_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: x_denominator(self,x),indices))))
        new_xgi=((self.prio_tau_x*self.prio_mu_x)+numerator_sum)/(self.prio_tau_x+denominator_sum)
        if self.debug:print(f"Old xgi {self.x[self.hash_x[gi]]}.",sys.stdout)
        self.x[self.hash_x[gi]]= new_xgi
        if self.debug: print(f"New xgi {self.x[self.hash_x[gi]]}.", sys.stdout)

def xx_loop(self):
    if self.verbose:
        print("Updating xx values",file=sys.stdout)
    for gi in self.unique_guides:
        indices = collect_all_keys(self, gi)
        numerator_sum = np.sum(np.asarray(list(map(lambda x: x_numerator(self, x), indices))))
        denominator_sum = np.sum(np.asarray(list(map(lambda x: x_denominator(self, x), indices))))
        new_xxgi = ((self.prio_tau_xx * self.prio_mu_xx) + numerator_sum) / (self.prio_tau_xx + denominator_sum)
        if self.debug: print(f"Old xxgi {self.xx[self.hash_x[gi]]}.", sys.stdout)
        self.xx[self.hash_x[gi]] = new_xxgi
        if self.debug: print(f"New xxgi {self.xx[self.hash_x[gi]]}.", sys.stdout)

def x_numerator(self,key_dict):
    numerator=(self.tau[self.hash_t[key_dict["t"]]]*self.y[self.hash_y[key_dict["gl"]]])*\
    (self.D[self.hash_d[key_dict["t"]]]-\
    (self.x[self.hash_x[key_dict["hj"]]]*self.y[self.hash_y[key_dict["hl"]]])-\
    (self.x[self.hash_x[key_dict["fk"]]]*self.y[self.hash_y[key_dict["fl"]]])-\
    (self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]))
    return numerator

def x_denominator(self,key_dict):
    denominator = self.tau[self.hash_t[key_dict["t"]]] * self.yy[self.hash_y[key_dict["gl"]]]
    return denominator


def collect_all_keys(self,gi_extern):
    #only get keys from tau and split them accordingly
    tau_keys = [x for x in self.hash_t.keys() if gi_extern in x] #TODO check in "if x.startswith(gi)" is needed
    out=[]
    for k in tau_keys:
        gi,hj,fk,l = k.split("_")
        gl = self.guide_to_gene(gi)
        hl = self.guide_to_gene(hj)
        fl = self.guide_to_gene(fk)
        s = find_s_keys(self,gl,hl,fl,l)
        r = find_r_keys(self,gi,hj,fk)
        out.append({"t":k,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s})
    return out




# import modules
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from sklearn.mixture import GaussianMixture
from scipy.stats import norm
import statsmodels.stats.multitest as multi
import os
import platform
if platform.system() == 'Windows':
    os.environ["MKL_NUM_THREADS"] = "1"
    os.environ["NUMEXPR_NUM_THREADS"] = "1"
    os.environ["OMP_NUM_THREADS"] = "1"                            # avoid warning message when using sklearn on Windows


class Score(object):
    """
    This class comprises functions for scoring triple gene interactions (strong interactions, lethality and recovery).
    """

    def __init__(self, model, nc_genes, pc_genes=None, pc_weight=0.5, verbose=False, output_dir=os.getcwd(),plotting=False):
        '''
        The inferred values from the class Model1 as well as positive and negative control genes are used as input.

        :param model: sample-independent effect (x) of guide i to gene g, sample-dependent individual effect (y),
         sample-dependent combinatorial effect (s), mean absolute error (MAE), and other parameters
        :param pc_genes: string or list of strings of positive control genes for removing interactions which
        involve individually lethal genes
        :param nc_genes: string or list of strings of negative control genes for building a Gaussian model
        :param pc_weight: weight of positive control gene
        :param verbose : sets verbosity, defaults to False
        :param output_dir : sets output directory, defaults to current working directory
        :param plotting : bool if plots should be generated, defaults to False
        :return: 4 values (three pandas dataframes with scores of triple gene interactions, cell line info)
        '''
        self.model = model                                                      # initialise output of class Model1
        self.verbose = verbose                                                  # initialise verbose for logging output
        self.plotting = plotting
        if os.path.exists(output_dir):                                          # check if supplied output dir exists
            self.output_dir = output_dir
        else:
            try:
                os.makedirs(output_dir) #try creating supplied output dir
                self.output_dir = output_dir
            except:
                print(f"Setting output directory to {os.getcwd()}")  # or use cwd as default
                self.output_dir = os.getcwd()
        if not self.model.is_inferred:
            raise TypeError("Run start_inference_rust() before calculating scores.")    # abort __init__
        elif self.model.is_inferred and self.verbose:                           # explicit opt-in for verbose output
            print("Model1 was successfully updated.")
        if pc_genes is None:
            self.pc_genes = []
        else:
            try:
                self.pc_genes =  self.check_input(pc_genes,"Positive control")   # check positive control genes
            except:
                raise
        if len(self.pc_genes) == 0:                                             # check if any pc genes were supplied
            print("No sensitive scores will be calculated, as no positive control gene was provided.")
        try:
            self.nc_genes = self.check_input(nc_genes, "Negative control")          # check negative control genes
        except:
            raise
        if pc_weight != 0.5:
            self.pc_weight = pc_weight                                          # initialise weight of pc gene
        else:
            self.pc_weight = pc_weight
            if self.verbose:
                print("Weight for positive control gene was set to 0.5 by default.")
        self.score_result = self.scoring()         # define output of class Score
        if self.plotting:
            self.plots = self.plotting_function()                                            # initialise drawing of plots


    def check_input(self, gene, name : str):
        #TODO are control genes in data?
        '''
        Function to check if positive and negative control genes were supplied.
        :param gene: name of variable of positive or negative control genes
        :param name: purpose of variable ("Positive control" or "Negative control")
        :return: If true: initialized variable, if false: error message
        '''
        if len(gene) == 0:                                                      # check if any nc genes were supplied
            raise TypeError(name + " genes must be supplied")
        elif isinstance(gene, str):                                        # type string and list are allowed
            self.gene = gene
        elif isinstance(gene, list):
            if all([isinstance(x,str) for x in gene]):
                self.gene = gene
        else:
            raise TypeError(name + " genes must be supplied as str or list of strings")
        return self.gene


    def new_dic(self, res, hashes):
        '''
        Function to modify the data structure.
        :param res: gene names
        :param hashes: inferred values
        :return: dictionary containing gene names as keys and inferred values as values
        '''
        #gene_list = [] if gene_list is None else gene_list
        #num_list = [] if num_list is None else num_list
        #val_list = [] if val_list is None else val_list
        dic = dict()

        for key, value in hashes.items():                                       # iterate through dictionary
            #gene_list.append(key)                                               # add genes as keys to list
            #num_list.append(value)                                              # add numbers of genes to list
            dic.update({key:res[value]})

        #for i in num_list:                                                      # i=0 or i=(0,0)
        #    val_list.append(res[i])                                             # add value of genes/guides to list

        #dic = dict(zip(gene_list, val_list))                                    # convert two np arrays to dict
        return dic


    def df_triple(self, gene_triple):
        '''
        Function to create a data frame where triple genes are mapped to their single genes.
        :param gene_triple: dictionary containing names of three genes as keys and inferred values as values
        :return: pandas dataframe with names of three genes as row names and three columns with single gene names
        '''
        df = pd.DataFrame.from_dict(gene_triple, orient='index')                # create df from dict using keys as rows
        df[["gene1", "gene2", "gene3"]] = ""                                    # add 3 new columns with empty str
        df = df.drop(0, axis=1)                                                 # delete column with s-values

        for rowname, col in df.iterrows():                                      # loop through rows and columns of df
            col[0] = rowname.split('_')[0]                                      # split row name by underscore
            col[1] = rowname.split('_')[1]                                      # and map 1st element to 1st col etc.
            col[2] = rowname.split('_')[2]

        # remove cell line info in row name & keep only unique rows (be independent from cell lines)
        df = df.rename(index=lambda x: x.rsplit('_', 1)[0]).drop_duplicates()
        return df


    def cell_info(self, y):
        '''
        Function to extract the information how many and which cell lines exist in data.
        :param y: dictionary containing name of one gene and cell line as keys and inferred values as values
        :return: list with names of cell lines as strings
        '''
        cell_list = []
        for key in y.keys():                                                               # iterate through names of keys
            cell = key.rsplit('_', 1)[1]                                            # keep only cell line info
            cell_list.append(cell)
        cell_list = list(set(cell_list))                                  # remove duplicate entries
        return cell_list


    def get_y(self, genes, y, cell_list):
        '''
        Function to create a numpy array containing the y values for calculating scores
        :param genes: names of genes at position 1, 2, or 3 (g,h,f)
        :param y: dictionary containing name of one gene and cell line as keys and inferred values as values
        :param cell_list: list with names of cell lines as strings
        :return: numpy array containing the y values of genes at position 1, 2, or 3 for calculating scores,
                 if multiple cell lines, the output is a stacked np array with one column per cell line
        '''
        lst = []

        if len(cell_list) == 1:
            for gene in genes:                                                  # iterate through names in array
                for key in sorted(y.keys()):                                                   # iterate through names of keys
                    key_n = key.rsplit('_', 1)[0]                               # remove info about cell line
                    if gene == key_n:                                           # if gene name matches key:
                        lst.append(y[key])                                      # - append list by value of key
            return np.expand_dims(np.array(lst), axis=1)                        # return array with y values

        elif len(cell_list) > 1:
            array = np.zeros((len(genes), 1))
            for i in cell_list:                                                 # iterate through cell lines
                y_cell = {k: v for k, v in y.items() if i in k}   # get dictionary with y values for cell line
                for gene in genes:
                    for key in sorted(y_cell.keys()):                                          # iterate through keys of cell line
                        key_n = key.rsplit('_', 1)[0]                               # remove info about cell line
                        if gene == key_n:                                       # if gene name matches key:
                            lst.append(y_cell[key])                             # - append list by value of key
                gene_values = np.expand_dims(np.array(lst), axis=1)          # add one dimension to reshape values array
                array = np.column_stack((array, gene_values))
                lst.clear()                                                     # empty list before starting loop again
            array = np.delete(array, 0, 1)                                      # delete first column containing zeros
            return array

        else:
            raise TypeError("Cell line information missing")                    # len(y) == 0

    def get_s(self,genes,s,cell_list):
        '''
        Function to create a numpy array containing the s values for calculating scores
        :param genes: names of genes in triple
        :param y: dictionary containing name of one triple and cell line as keys and inferred values as values
        :param cell_list: list with names of cell lines as strings
        :return: numpy array containing the s values of genes  for calculating scores,
                 if multiple cell lines, the output is a stacked np array with one column per cell line
        '''
        lst = []

        if len(cell_list) == 1:
            for gene in genes:  # iterate through names in array
                for key in sorted(s.keys()):  # iterate through names of keys
                    key_n = key.rsplit('_', 1)[0]  # remove info about cell line
                    if gene == key_n:  # if gene name matches key:
                        lst.append(s[key])  # - append list by value of key
            return np.expand_dims(np.array(lst), axis=1)  # return array with y values

        elif len(cell_list) > 1:
            array = np.zeros((len(genes), 1))
            for i in cell_list:  # iterate through cell lines
                s_cell = {k: v for k, v in s.items() if i in k}  # get dictionary with y values for cell line
                for gene in genes:
                    for key in sorted(s_cell.keys()):  # iterate through keys of cell line
                        key_n = key.rsplit('_', 1)[0]  # remove info about cell line
                        if gene == key_n:  # if gene name matches key:
                            lst.append(s_cell[key])  # - append list by value of key
                gene_values = np.expand_dims(np.array(lst), axis=1)  # add one dimension to reshape values array
                array = np.column_stack((array, gene_values))
                lst.clear()  # empty list before starting loop again
            array = np.delete(array, 0, 1)  # delete first column containing zeros
            return array

        else:
            raise TypeError("Cell line information missing")

    def calculate_scores(self, yg, yh, yf, sghf, mod_y):
        '''
        Function to compute the scores strong, sensitive lethality and sensitive recovery.
        :param yg: numpy array containing the y values of genes at position 1 (for every cell line)
        :param yh: numpy array containing the y values of genes at position 2 (for every cell line)
        :param yf: numpy array containing the y values of genes at position 3 (for every cell line)
        :param sghf: inferred values for triple gene interactions
        :param mod_y: dictionary containing name of one gene and cell line as keys and inferred values as values
        :return: numpy array with scores in correct order
        '''
        strong = np.abs(sghf) - np.abs(yg) - np.abs(yh) - np.abs(yf)            # Calculate strong score

        # Sensitive scores are only calculated if positive control gene was provided
        if len(self.pc_genes) > 0:
            sensitive_lethality_a = np.amin([yg, yh, yf],axis=0) - yg - yh - yf - sghf
            sensitive_recovery_a = yg + yh + yf + sghf - np.amin([yg, yh, yf],axis=0)

            # Update sensitive scores because one gene can either have a lethality or recovery effect
            pc_gene_keylist = [x for x in mod_y if x.split('_')[0] in self.pc_genes]
            pc_values = [mod_y[i] for i in pc_gene_keylist]
            pc_gene_mean = np.asarray(pc_values).mean()                         # compute mean of cell lines for pc gene
            threshold = pc_gene_mean * self.pc_weight

            if threshold >= 0:
                # Lethality: keep pairs if yg & yh & yf > threshold (if true = 1, if false = 0)
                lethality = np.where((yg >= threshold) & (yh >= threshold) & (yf >= threshold), 1, 0)
                # Recovery: keep pairs if one gene is smaller than threshold
                recovery = np.where((yg < threshold) | (yh < threshold) | (yf < threshold), 1, 0)
            else:
                lethality = np.where((yg < threshold) & (yh < threshold) & (yf < threshold), 1, 0)
                recovery = np.where((yg >= threshold) | (yh >= threshold) | (yf >= threshold), 1, 0)

            # Filter scores with mask lethality / recovery
            sensitive_lethality = np.where((lethality == 1), sensitive_lethality_a, np.nan)
            sensitive_recovery = np.where((recovery == 1), sensitive_recovery_a, np.nan)

        # Create sensitive scores with same shape as strong score but filled with zeros
        else:
            sensitive_lethality = np.zeros(np.shape(strong))
            sensitive_recovery = np.zeros(np.shape(strong))

        return strong, sensitive_lethality, sensitive_recovery


    def nc_values(self, keys_tgenes):
        '''
        Function to define scores including negative control genes
        :param keys_tgenes: Numpy array with names of all triple gene combinations
        :return: np array containing names of triple gene combinations which include at least one nc gene
        '''
        if isinstance(self.nc_genes, type('')):                                             # check if type str
            ncgenes_values = np.array([x for x in keys_tgenes if self.nc_genes in x])
        elif isinstance(self.nc_genes, type([])):                                           # check if type list
            ncgenes_all = []
            for index, gene in enumerate(self.nc_genes):
                ncgene = [x for x in keys_tgenes if self.nc_genes[index] in x]
                ncgenes_all = ncgenes_all + ncgene
            ncgenes_values = np.array(list(dict.fromkeys(ncgenes_all)))                     # remove duplicate entries
        else:
            raise TypeError("Negative control genes must be supplied as str or list of strings.")
        return ncgenes_values


    def prepare_nc_model(self, keys_tgenes, strong, sens_lethality, sens_recovery, nc_values, cell):
        '''
        Function to define a negative control model
        :param keys_tgenes: Numpy array with names of all triple gene combinations
        :param strong: Numpy array containing strong scores
        :param sens_lethality: Numpy array containing sensitive lethality scores
        :param sens_recovery: Numpy array containing sensitive recovery scores
        :param nc_values: np array containing names of triple gene combinations which include at least one nc gene
        :param cell: Integer indicating the cell line
        :return: 3 np arrays containing the scores of the negative control combinations
                 (for strong, sensitive lethality and sensitive recovery scores)
        '''
        strong_dic = dict(zip(keys_tgenes, strong[:, cell]))         # create new dictionary for score values
        strong_nc = np.array([strong_dic[k] for k in nc_values])     # obtain only values whose keys are nc combinations

        if len(self.pc_genes) > 0:
            lethality_dic = dict(zip(keys_tgenes, sens_lethality[:, cell]))
            recovery_dic = dict(zip(keys_tgenes, sens_recovery[:, cell]))
            lethality_nc = np.array([lethality_dic[k] for k in nc_values])
            lethality_nc = lethality_nc[~np.isnan(lethality_nc)]                                # drop nan
            recovery_nc = np.array([recovery_dic[k] for k in nc_values])
            recovery_nc = recovery_nc[~np.isnan(recovery_nc)]                                   # drop nan
        else:
            lethality_nc = np.zeros(np.shape(strong_nc))
            recovery_nc = np.zeros(np.shape(strong_nc))

        return strong_nc, lethality_nc, recovery_nc


    def gmm(self, nc_input, title, cell):
        '''
        Function to find a Gaussian mixture model (GMM) which best describes the distribution of negative control scores.
        sklearn.mixture and GMMs with 1-5 Gaussians are used.
        :param nc_input: Numpy array containing the scores of the negative control combinations
        :param title: 'Strong', 'lethality' or 'recovery'
        :param cell: Name of cell line as string
        :return: GMM which was considered as best solution by the Bayesian information criterion.
        '''
        nc = np.reshape(nc_input, (-1, 1))                                      # reshaping required for fit

        # fit Gaussian Mixture Models with 1 to max. 5 components, respectively, and find the best fit
        if np.size(nc) < 5:
            number = np.arange(1, np.size(nc) + 1)                              # array with length of nc array
        else:
            number = np.arange(1, 6)                                            # array with 5 elements
        models = [None for i in range(len(number))]                             # list with 5 entries "None"

        for i in range(len(number)):
            models[i] = GaussianMixture(n_components=number[i], tol=0.0001, max_iter=100, n_init=3,
                                        init_params='kmeans', verbose=0).fit(nc)

        bic = [m.bic(nc) for m in models]                                       # Bayesian information criterion
        best_model = models[bic.index(min(bic))]                                # Minimum of BIC defines best model

        # Print information about current status in terminal
        if models[bic.index(min(bic))].converged_ is True and self.verbose:
            print("Gaussian Mixture Model converged.")
        else:
            print("No convergence was reached for Gaussian Mixture Model.")
        txt = "According to the Bayesian information criterion, {0} steps were required" \
              " by the best fit of EM to converge."
        if self.verbose:
            print(txt.format(models[bic.index(min(bic))].n_iter_))
        if np.argmin(bic) == 0:
            print("Best GMM consists of 1 Gaussian.")
        else:
            print("Best GMM consists of", np.argmin(bic)+1, "Gaussians.")

        if self.plotting:
            # Plot figures showing the result of fitting GMM to the data
            fig = plt.figure(figsize=(12, 8))
            fig.subplots_adjust(left=0.12, right=0.9, bottom=0.21, top=0.9, wspace=0.5)

            # plot 1: data + best-fit mixture
            ax = fig.add_subplot(121)                                               # subplot 1 (left)
            m_best = models[np.argmin(bic)]
            lin = np.linspace(min(nc) - 0.1, max(nc) + 0.1, num=100)
            logprob = m_best.score_samples(lin.reshape(-1, 1))
            responsibilities = m_best.predict_proba(lin.reshape(-1, 1))
            pdf = np.exp(logprob)                                                   # probability density
            pdf_individual = responsibilities * pdf[:, np.newaxis]
            ax.hist(nc, 30, density=True, histtype='stepfilled', color='#009FE3', alpha=0.4)
            if np.argmin(bic) == 0:                                                 # GMM with only one component
                ax.plot(lin, pdf, '--k')                                            # display model without legend
            else:                                                                   # GMMs with 2 - 5 components
                def legend(ax):                                                     # legend without redundant labels
                    handles, label = ax.get_legend_handles_labels()
                    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, label)) if l not in label[:i]]
                    ax.legend(*zip(*unique), loc=2, fontsize=14)
                ax.plot(lin, pdf, '--k', label='Gaussian mixture')
                ax.plot(lin, pdf_individual, ':k', label='individual Gaussians')
                legend(ax)
            ax.text(0.0, 1.08, "Best-fit GMM " + title + cell, ha='left', va='top', transform=ax.transAxes, fontsize=20)
            ax.set_xlabel('score nc gene combinations', fontsize=18)
            ax.set_ylabel('count', fontsize=20)
            ax.tick_params(axis='both', labelsize=14)

            # plot 2: BIC
            ax = fig.add_subplot(122)                                               # subplot 2 (right)
            ax.plot(number, bic, color='#009FE3', label='BIC')
            ax.set_xticks(np.arange(min(number), max(number) + 1, 1.0))             # define step size of x ticks
            ax.text(0.0, 1.08, 'Number of Gaussians ' + cell, ha='left', va='top', transform=ax.transAxes, fontsize=20)
            ax.set_xlabel('components', fontsize=20)
            ax.set_ylabel('information criterion', fontsize=20)
            ax.tick_params(axis='both', labelsize=14)
            ax.legend(loc=2, fontsize=16)                                           # legend on the top left

            plt.savefig(os.path.join(self.output_dir, "Output", "GMM", ("GMM_" + title + "_" + cell + ".png")))
            plt.close()
        return best_model


    def p_value(self, nc_model, score):
        '''
        Function for calculating P-values for each individual score. A P-value describes the probability of the test
        result given an actual distribution of negative controls (approximated by the Gaussian mixture model).
        p_values = 1 - (weight * (norm(mean, stDev).cdf(score)))
        :param nc_model: Gaussian mixture models which best describes the distribution of nc genes
        :param score: array of all 'strong', 'lethality' or 'recovery' scores, respectively
        :return: Numpy array of P-values in same order as scores
        '''
        n = np.size(nc_model.means_)                                            # get total number of elements in array
        added_elements = 0
        for i in range(0, n, 1):
            element = nc_model.weights_[i] * (norm(nc_model.means_[i], np.sqrt(nc_model.covariances_[i])).sf(score))
            added_elements = added_elements + element
        pvalue = added_elements
        return np.ravel(pvalue)


    def fdr(self, pvalue):
        '''
        Function for calculating adjusted P-values (FDR) by using the method of Benjamini and Hochberg.
        P-values are sorted before calculating the FDRs.
        :param pvalue: Numpy array of P-values of one score
        :return: Numpy array of adjusted P-values of one score
        '''
        pvalue_sort = np.sort(pvalue, axis=None)                                # returns the flattened array, sorted
        pvalue_argsort = np.argsort(pvalue, axis=None)                          # returns original indices of array
        mask = np.isfinite(pvalue_sort)                                         # mask for pvalues which are not NAN

        # Benjamini/Hochberg (non-negative)
        fdr_bh = multi.multipletests(pvalue_sort[mask], alpha=0.05, method='fdr_bh', is_sorted=True)

        fdr_bh_sorted = np.empty_like(pvalue)  # empty array to put back in the NAN
        fdr_bh_sorted_out = np.empty_like(pvalue)  # empty array to return
        fdr_bh_sorted.fill(np.nan)
        fdr_bh_sorted[mask] = fdr_bh[1]

        fdr_bh_sorted_out[pvalue_argsort] = fdr_bh_sorted  # return fdr to original order
        return fdr_bh_sorted_out


    def dataframe(self, score, pvalue, fdr, cell_list, line, names):
        '''
        Function for creating an empty output data frame.
        :param score: array of all 'strong', 'lethality' or 'recovery' scores, respectively
        :param pvalue: Numpy array of P-values
        :param fdr: Numpy array of adjusted P-values
        :param cell_list: String or list of cell lines included in CRISPR screen
        :param line: Integer indicating the cell line
        :param names: Data frame where triple genes are mapped to their single genes
        :return: Data frame with three columns (Score, P-value, Adj_P-value) for the first cell line
        '''
        df = pd.DataFrame(data=np.column_stack((score, pvalue, fdr)), index=names.index.tolist(),
                          columns=['Score_' + cell_list[line],
                                   'P-value_' + cell_list[line],
                                   'Adj_P-value_' + cell_list[line]])
        return df


    def add_columns(self, df, score, pvalue, fdr, cell_list, line):
        '''
        Function for adding three columns to the output data frame.
        :param df: Data frame with three columns (Score, P-value, Adj_P-value) for every cell line
        :param score: array of all 'strong', 'lethality' or 'recovery' scores, respectively
        :param pvalue: Numpy array of P-values
        :param fdr: Numpy array of adjusted P-values
        :param cell_list: String or list of cell lines included in CRISPR screen
        :param line: Integer indicating the cell line
        :return: Data frame with three new columns for every additional cell line
        '''
        df['Score_' + cell_list[line]] = score[:, line].tolist()
        df['P-value_' + cell_list[line]] = pvalue.tolist()
        df['Adj_P-value_' + cell_list[line]] = fdr.tolist()
        return df


    def get_top10_df(self, df, cellpos, cellline):
        '''
        Function for creating a data frame for the 10 gene triples with the strongest effects.
        :param df: Output data frame of the strong score
        :param cellpos: Integer indicating the cell line
        :param cellline: Name of the cell line as string
        :return: Data frame with 10 rows for the most interesting triple combinations with a lethal, recovering,
                 or no effect (directly saved in output folder)
        '''
        if df.iloc[:, cellpos * 4].str.contains('lethality').any():
            lethality_df = df[df.iloc[:, cellpos * 4] == 'lethality'].iloc[:, cellpos * 4:cellpos * 4 + 4]
            top10_lethal = lethality_df.sort_values(by=[lethality_df.columns[1]], ascending=False).iloc[0:10, :]
            top10_lethal.to_csv(os.path.join(self.output_dir, "Output", ("top10_lethal_" + cellline + ".csv")))
            top10_lethal.name = 'top10_lethal_' + cellline
            self.scatter(top10_lethal)

        if df.iloc[:, cellpos * 4].str.contains('recovery').any():
            recovery_df = df[df.iloc[:, cellpos * 4] == 'recovery'].iloc[:, cellpos * 4:cellpos * 4 + 4]
            top10_recovery = recovery_df.sort_values(by=[recovery_df.columns[1]], ascending=False).iloc[0:10, :]
            top10_recovery.to_csv(os.path.join(self.output_dir, "Output", ("top10_recovery_" + cellline + ".csv")))
            top10_recovery.name = 'top10_recovery_' + cellline
            self.scatter(top10_recovery)

        if df.iloc[:, cellpos * 4].str.contains('no').any():
            noeffect_df = df[df.iloc[:, cellpos * 4] == 'no effect'].iloc[:, cellpos * 4:cellpos * 4 + 4]
            top10_noeffect = noeffect_df.sort_values(by=[noeffect_df.columns[1]], ascending=False).iloc[0:10, :]
            top10_noeffect.to_csv(os.path.join(self.output_dir, "Output", ("top10_noeffect_" + cellline + ".csv")))
            top10_noeffect.name = 'top10_noeffect_' + cellline
            self.scatter(top10_noeffect)


    def scoring(self):
        '''
        Main function of the class Score.
        All the other functions described above are called from this function.
        :return: - 3 final data frames, one for each score (columns for effect, score, P-value, and FDR per cell line)
                 - Cell lines included in CRISPR screen (as integer or list)
        '''
        # Create new directory for outputs
        if not os.path.exists(os.path.join(self.output_dir, "Output")):
            os.makedirs(os.path.join(self.output_dir, "Output"))

        # Create new directory for results and plots of Gaussian Mixture Model
        if not os.path.exists(os.path.join(self.output_dir, "Output", "GMM")):
            os.makedirs(os.path.join(self.output_dir, "Output", "GMM"))

        # Create new directory for plots
        if not os.path.exists(os.path.join(self.output_dir, "Output", "Plots")):
            os.makedirs(os.path.join(self.output_dir, "Output", "Plots"))

        # Modify structure of output data from of Model1 (create informative dictionaries)
        mod_y = self.new_dic(self.model.y, self.model.hash_y)
        mod_s = self.new_dic(self.model.s, self.model.hash_s)

        # Create data frame where triple genes are mapped to their single genes
        Tgenes2Sgenes = self.df_triple(mod_s)
        keys_tgenes = Tgenes2Sgenes.index.to_numpy()                                      # convert rownames to np array

        # Get information about which cell lines are present in data
        cell_list = self.cell_info(mod_y)

        # Create arrays with y values in correct order
        yg = self.get_y(Tgenes2Sgenes.iloc[:, 0].values, mod_y, cell_list)                # use columns of Tgenes2Sgenes
        yh = self.get_y(Tgenes2Sgenes.iloc[:, 1].values, mod_y, cell_list)                # and values from mod_y array
        yf = self.get_y(Tgenes2Sgenes.iloc[:, 2].values, mod_y, cell_list)
        sghf = self.get_s(Tgenes2Sgenes.index,mod_s,cell_list)

        # Compute scores
        strong, sensitive_lethality, sensitive_recovery = self.calculate_scores(yg, yh, yf, sghf, mod_y)

        # Create np array with values of negative control genes
        ncgenes_values = self.nc_values(keys_tgenes)

        # Create nc_model per cell line to calculate p-value and FDR per cell line
        for cell_line in range(len(cell_list)):

            # Obtain negative control gene values within each score (as input for Gaussian Mixture Model)
            strong_nc, lethality_nc, recovery_nc = self.prepare_nc_model(keys_tgenes, strong, sensitive_lethality,
                                                                         sensitive_recovery, ncgenes_values, cell_line)

            # Check distribution of negative control gene score values before running EM algorithm
            if self.plotting:
                self.nc_distribution(strong_nc, lethality_nc, recovery_nc, cell_list[cell_line])

            # Running expectation maximization algorithm from Gaussian Mixture on the three scores
            strong_nc_model = self.gmm(strong_nc, "strong ", cell_list[cell_line])
            if len(self.pc_genes) > 0:
                sensitive_lethality_nc_model = self.gmm(lethality_nc, "lethality ", cell_list[cell_line])
                sensitive_recovery_nc_model = self.gmm(recovery_nc, "recovery ", cell_list[cell_line])

            # Calculate P-values (null hypothesis: strong_nc_model > strong)
            pvalue_strong = self.p_value(strong_nc_model, strong[:, cell_line])
            if len(self.pc_genes) > 0:
                pvalue_lethality = self.p_value(sensitive_lethality_nc_model, sensitive_lethality[:, cell_line])
                pvalue_recovery = self.p_value(sensitive_recovery_nc_model, sensitive_recovery[:, cell_line])

            # Calculate Adjusted P-values (FDR)
            fdr_strong = self.fdr(pvalue_strong)
            if len(self.pc_genes) > 0:
                fdr_lethality = self.fdr(pvalue_lethality)
                fdr_recovery = self.fdr(pvalue_recovery)

            # Extract information whether strong combinatorial effect is lethal or recovering
            effect_type = []
            for i in range(len(strong[:, cell_line])):                                  # iterate through rows of strong
                if strong[i, cell_line] > 0 and sghf[i, cell_line] >= 0:
                    effect_type.append('recovery')
                elif strong[i, cell_line] > 0 and sghf[i, cell_line] < 0:
                    effect_type.append('lethality')
                else:
                    effect_type.append('no effect')
            strong_type = np.array(effect_type)

            # Create output df (with initially three columns, and 3 extra columns per additional cell line)
            if cell_line == 0:
                strong_df = self.dataframe(strong[:, cell_line], pvalue_strong, fdr_strong,
                                           cell_list, cell_line, Tgenes2Sgenes)
                strong_df.insert(loc=0, column='Interaction_' + cell_list[cell_line], value=strong_type)
                if len(self.pc_genes) > 0:
                    sensitive_lethality_df = self.dataframe(sensitive_lethality[:, cell_line], pvalue_lethality,
                                                            fdr_lethality, cell_list, cell_line, Tgenes2Sgenes)
                    sensitive_recovery_df = self.dataframe(sensitive_recovery[:, cell_line], pvalue_recovery,
                                                           fdr_recovery, cell_list, cell_line, Tgenes2Sgenes)
                else:  # create empty data frames (required for order of elements returned by function "scoring()"
                    data = {'Score_' + cell_list[cell_line]: [0],
                            'P-value_' + cell_list[cell_line]: [0],
                            'Adj_P-value_' + cell_list[cell_line]: [0]}
                    sensitive_lethality_df = pd.DataFrame(data, index=['no data available'])
                    sensitive_recovery_df = pd.DataFrame(data, index=['no data available'])

            else:
                strong_df.insert(loc=cell_line*3+1, column='Interaction_' + cell_list[cell_line], value=strong_type)
                strong_df = self.add_columns(strong_df, strong, pvalue_strong, fdr_strong, cell_list, cell_line)
                if len(self.pc_genes) > 0:
                    sensitive_lethality_df = self.add_columns(sensitive_lethality_df, sensitive_lethality,
                                                              pvalue_lethality, fdr_lethality, cell_list, cell_line)
                    sensitive_recovery_df = self.add_columns(sensitive_recovery_df, sensitive_recovery,
                                                             pvalue_recovery, fdr_recovery, cell_list, cell_line)

            # Extract important information from output dfs
            self.get_top10_df(strong_df, cell_line, cell_list[cell_line])

        # Save output data frames
        strong_df.to_csv(os.path.join(self.output_dir, "Output", "strong_df.csv"))
        if len(self.pc_genes) > 0:
            sensitive_lethality_df.to_csv(os.path.join(self.output_dir, "Output", "sensitive_lethality_df.csv"))
            sensitive_recovery_df.to_csv(os.path.join(self.output_dir, "Output", "sensitive_recovery_df.csv"))

        print("Scores were successfully calculated.")
        return strong_df, sensitive_lethality_df, sensitive_recovery_df, cell_list


    # The second part of the class Score contains functions for creating visualizations
    def convergence(self, mae):
        '''
        Function to create a line plot showing the convergence of the CAVI algorithm.
        If model is divergent, alternative parameters for the priors should be selected until convergence is achieved.
        :param mae: Mean absolute error after every iteration (~ convergence rate).
        :return: Figure (saved in output folder)
        '''
        plt.figure(figsize=(12, 8))
        plt.plot(mae, color='#009FE3')
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        plt.title('Convergence of interference', fontsize=22)
        plt.xlabel('Iteration steps', fontsize=20)
        plt.ylabel('Mean absolute error', fontsize=20)
        plt.savefig(os.path.join(self.output_dir, "Output", "GMM", "Convergence.png"))
        plt.close()


    def nc_distribution(self, strong_nc, sens_lethality_nc, sens_recovery_nc, cell):
        '''
        Function to depict the distribution of negative control genes (should be around 0 = no effect).
        :param strong_nc: Numpy array of strong scores containing negative control genes.
        :param sens_lethality_nc: Numpy array of sensitive lethality scores containing negative control genes.
        :param sens_recovery_nc: Numpy array of sensitive recovery scores containing negative control genes.
        :param cell: Name of cell line as string.
        :return: Figure showing three box plots (saved in output folder)
        '''
        plt.figure(figsize=(12, 8))

        if len(self.pc_genes) > 0:
            box = plt.boxplot([strong_nc, sens_lethality_nc, sens_recovery_nc], patch_artist=True, showmeans=True)
            plt.setp(box['means'], marker='D')
            c = ['#0047B9', '#FF5C00', '#00CC00']
            c2 = ['#0047B9', '#0047B9', '#FF5C00', '#FF5C00', '#00CC00', '#00CC00']
            c3 = ['#009FE3', '#C61826', 'darkgreen']
            ticks = [1, 2, 3]
            labels = ['strong', 'sensitive\nlethality', 'sensitive\nrecovery']
        else:
            box = plt.boxplot([strong_nc], patch_artist=True, showmeans=True)
            plt.setp(box['means'], marker='D')
            c = ['#0047B9']
            c2 = ['#0047B9', '#0047B9']
            c3 = ['#009FE3']
            ticks = [1]
            labels = ['strong']

        # Define colors of box plots
        for patch, color in zip(box['boxes'], c):
            patch.set_color(color)
            patch.set_facecolor(color)
        for patch, color in zip(box['fliers'], c):
            patch.set_color(color)
            patch.set_markeredgecolor(color)
        for patch, color in zip(box['whiskers'], c2):
            patch.set_color(color)
        for patch, color in zip(box['caps'], c2):
            patch.set_color(color)
        for patch, color in zip(box['means'], c3):
            patch.set_markeredgecolor(color)
            patch.set_markerfacecolor(color)
        for patch, color in zip(box['medians'], c3):
            patch.set_color(color)

        plt.xticks(ticks, labels, fontsize=20)
        plt.yticks(fontsize=14)
        plt.ylabel('scores', fontsize=20)
        title = 'Distribution of negative control genes ' + cell
        plt.title(title, fontsize=20)
        plt.legend([box['medians'][0], box['means'][0]], ['median', 'mean'], fontsize=16)
        title = title.replace(" ", "_")
        plt.savefig(os.path.join(self.output_dir, "Output", "GMM", (title+".png")))
        plt.close()


    def LFC_distribution(self):
        '''
        Function to show the distribution of log2 fold changes.
        :return: Box plot of LFCs (saved in output folder)
        '''
        plt.figure(figsize=(12, 8))
        c = "#0047B9"
        box = plt.boxplot(self.model.D, patch_artist=True, showmeans=True,
                          boxprops=dict(facecolor=c, color=c),
                          capprops=dict(color=c),
                          whiskerprops=dict(color=c),
                          flierprops=dict(color=c, markeredgecolor=c),
                          medianprops=dict(color='#009FE3'),
                          meanprops=dict(markeredgecolor='#009FE3', markerfacecolor='#009FE3', marker='D'))
        ticks = list(range(len(self.score_result[3]) + 1))                             # create 1 tick per cell line
        ticks.pop(0)                                                                   # delete first element which is 0
        plt.xticks(ticks, self.score_result[3], fontsize=18)
        plt.ylabel('$Log_2$ fold changes', fontsize=20)
        plt.yticks(fontsize=14)
        title = 'Distribution of LFC'
        plt.title(title, fontsize=20)
        plt.legend([box['medians'][0], box['means'][0]], ['median', 'mean'], fontsize=16)
        plt.savefig(os.path.join(self.output_dir, "Output", "Plots", (title + ".png")))
        plt.close()


    def scatter(self, top10df):
        '''
        Function to create a scatter plot showing the 10 gene combinations with the strongest effect.
        :param top10df: Data frame with 10 rows for the most lethal or recovering triple combinations
        :return: Scatter plot (saved in output folder)
        '''
        effect = top10df.name.split('_')[1]                           # effect = 'lethality', 'recovery', or 'no effect'
        cell = top10df.name.split('_')[2]                                               # get cell line
        y = top10df.iloc[:, 1]
        x = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        x = x[0:np.size(y)]                                                             # adjust length of x by y
        if effect == 'lethal':                                                          # orange color palette
            colors = ['#A83C00', '#BF4500', '#F25600', '#FF5C00', '#FF9053',
                      '#FF9D66', '#FFA675', '#FFBE99', '#FFB78F', '#FFDECC']
        elif effect == 'recovery':                                                      # green color palette
            colors = ['#008600', '#009E00', '#00B000', '#00C000', '#00C800',
                      '#05FF05', '#3FFF3F', '#5DFF5D', '#89FF89', '#A7FFA7']
        else:                                                                           # blue color palette (no effect)
            colors = ['#006D9E', '#0077AA', '#0094D6', '#009FE3', '#00AAF6',
                      '#47C6FF', '#55CCFF', '#75D4FF', '#8EDDFF', '#A7E4FF']
        plt.figure(figsize=(12, 8))
        plt.scatter(x, y, s=350, c=colors[0:np.size(y)])
        plt.xticks(x, top10df.index, rotation=45, fontsize=14)                          # set text labels and properties
        plt.yticks(fontsize=14)
        plt.margins(0.1)
        plt.subplots_adjust(bottom=0.35)
        plt.ylabel('score', fontsize=20)
        if effect == 'noeffect':
            title = 'Top ' + str(np.size(y)) + ' triple gene interactions with no effect ' + cell
        else:
            title = 'Top ' + str(np.size(y)) + ' ' + effect + ' triple gene interactions ' + cell
        plt.title(title, fontsize=20)
        title = title.replace(" ", "_")
        plt.savefig(os.path.join(self.output_dir, "Output", "Plots", (title+".png")))
        plt.close()


    def volcano_score(self, df, effect, col=None):
        '''
        Function to create a volcano plot showing the distribution and highlighting interesting calculated scores.
        :param df: Final data frame for score with the columns 'effect', 'score', 'P-value', and 'FDR' per cell line
        :param effect: String describing effect ('strong', 'sensitive lethality', or 'sensitive recovery')
        :return: Volcano plot showing distribution of scores (saved in output folder)
        '''
        col = [] if col is None else col
        if effect == 'strong':
            x = df.iloc[:, 1]
            y = (df.iloc[:, 1]) ** 2
            for i in (range(len(x))):
                col.append(np.where(df.iloc[i, 0] == 'lethality', '#FF5C00',
                                    np.where(df.iloc[i, 0] == 'recovery', '#00CC00', '#009FE3')))
        elif effect == 'sensitive lethality':
            x = df.iloc[:, 0]
            y = (df.iloc[:, 0]) ** 2
            for i in (range(len(x))):
                col.append(np.where(pd.isna(df.iloc[i, 0]), '#009FE3', '#FF5C00'))
        else:
            x = df.iloc[:, 0]
            y = (df.iloc[:, 0]) ** 2
            for i in (range(len(x))):
                col.append(np.where(pd.isna(df.iloc[i, 0]), '#009FE3', '#00CC00'))
        color = np.array(col)
        plt.figure(figsize=(12, 8))
        title = 'Distribution of ' + effect + ' scores'
        plt.title(title, fontsize=21)
        plt.xlabel('score', fontsize=20)
        plt.xticks(fontsize=14)
        plt.ylabel('$score^2$', fontsize=20)
        plt.yticks(fontsize=14)
        plt.scatter(x, y, s=120, c=color, alpha=0.6, edgecolors='none')                     # alpha: transparency
        a = mpatches.Patch(color='#FF5C00', label='lethality')
        b = mpatches.Patch(color='#00CC00', label='recovery')
        c = mpatches.Patch(color='#009FE3', label='no effect')
        plt.legend(handles=[a, b, c], fontsize=16)
        title = title.replace(" ", "_")
        plt.savefig(os.path.join(self.output_dir, "Output", "Plots", (title+".png")))
        plt.close()

    def volcano_LFC(self, df, col=None):
        '''
        Function to create a volcano plot showing the mean combinatorial gene effect vs -log10(P-value).
        In this way the effect sizes of the triple gene combinations can be compared.
        :param df: Final data frame for score with the columns 'effect', 'score', 'P-value', and 'FDR' per cell line
        :return: Volcano plot showing LFCs (saved in output folder)
        '''
        col = [] if col is None else col
        x = self.model.s.mean(axis=1)                                                               # mean log2 FC
        y = np.abs(np.log10(df.iloc[:, 2]))                                                         # -log10(P-value)
        for i in (range(len(x))):
            col.append(np.where(df.iloc[i, 0] == 'lethality', '#FF5C00',
                                np.where(df.iloc[i, 0] == 'recovery', '#00CC00', '#009FE3')))
        color = np.array(col)
        plt.figure(figsize=(12, 8))
        plt.scatter(x, y, s=120, c=color, alpha=0.6, edgecolors='none')                     # alpha: transparency
        title = 'Volcano strong combinatorial effect'
        plt.title(title, fontsize=21)
        plt.xlabel('Mean combinatorial effect', fontsize=20)
        plt.xticks(fontsize=14)
        plt.ylabel('$-log_{10}$(P-value)', fontsize=20)
        plt.yticks(fontsize=14)
        a = mpatches.Patch(color='#FF5C00', label='lethality')
        b = mpatches.Patch(color='#00CC00', label='recovery')
        c = mpatches.Patch(color='#009FE3', label='no effect')
        plt.legend(handles=[a, b, c], fontsize=14)
        plt.hlines(1.3, min(x), max(x), color='grey', linestyle='dashed')                   # plotting line at p=0.05
        plt.text(max(x), 1.3, 'p<0.05', fontsize=11, color='grey',
                 va='center_baseline', ha='center', backgroundcolor='w')

        # Create labels for the 2 most important lethal as well as recovering interactions
        pvalue = df.iloc[:, 2].to_numpy()                                            # extract P-values in correct order
        top2_lethality = df.sort_values(by=[df.columns[0], df.columns[1]], ascending=[True, False]).iloc[0:2, :]
        top2_recovery = df.sort_values(by=[df.columns[0], df.columns[1]], ascending=[False, False]).iloc[0:2, :]
        top2_lethality = top2_lethality[top2_lethality.iloc[:, 0] == 'lethality']    # keep only lethal interactions
        top2_recovery = top2_recovery[top2_recovery.iloc[:, 0] == 'recovery']        # keep only recovery interactions
        top2_lethality_np = top2_lethality.iloc[:, 2].to_numpy()                     # convert P-values to np array
        top2_recovery_np = top2_recovery.iloc[:, 2].to_numpy()

        pos_list_lethal = []
        pos_list_recovery = []
        for pos, value in enumerate(pvalue):
            for val_l in top2_lethality_np:
                if value == val_l:
                    pos_list_lethal.append(pos)
            for val_r in top2_recovery_np:
                if value == val_r:
                    pos_list_recovery.append(pos)

        # extract rownames for annotation of data points
        rownames = list(df.index)
        if pos_list_lethal == pos_list_recovery:                                    # in case there is only 'no effect'
            for pos in pos_list_lethal:
                plt.annotate(rownames[pos], xy=(x[pos], y[pos]), fontsize=12,
                             arrowprops=dict(fc='#009FE3', width=0.9, shrink=0.05),
                             horizontalalignment='center', verticalalignment='top',
                             bbox=dict(fc='#FF5C00', alpha=0.05, ec='#009FE3'))
        else:
            for pos in pos_list_lethal:
                plt.annotate(rownames[pos], xy=(x[pos], y[pos]), fontsize=12,
                             arrowprops=dict(fc='#FF5C00', width=0.9, shrink=0.05),
                             horizontalalignment='left', verticalalignment='top',
                             bbox=dict(fc='#FF5C00', alpha=0.05, ec='#FF5C00'))

            for pos in pos_list_recovery:
                plt.annotate(rownames[pos], xy=(x[pos], y[pos]), fontsize=12,
                             arrowprops=dict(fc='#00CC00', width=0.9, shrink=0.05),
                             horizontalalignment='right', verticalalignment='top',
                             bbox=dict(fc='#00CC00', alpha=0.05, ec='#00CC00'))

        title = title.replace(" ", "_")
        plt.savefig(os.path.join(self.output_dir, "Output", "Plots", (title+".png")))
        plt.close()


    def plotting_function(self):
        '''
        Main function to call all functions that create a figure.
        '''
        # Mean absolute error
        self.convergence(self.model.MAE)

        # Distribution log2 fold changes of triple gene interactions
        self.LFC_distribution()

        # Volcano plot LFC
        self.volcano_LFC(self.score_result[0])

        # Volcano plot with scores
        self.volcano_score(self.score_result[0], 'strong')
        if len(self.pc_genes) > 0:
            self.volcano_score(self.score_result[1], 'sensitive lethality')
            self.volcano_score(self.score_result[2], 'sensitive recovery')

        print("Plots were successfully drawn.\n"
              "Figures and resulting data frames can be seen in folder 'Output'.")



import numpy as np
import sys
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys

def r_loop(self):
    if self.verbose:
        print("Updating r values",file=sys.stdout)
    for r_triple in self.hash_r.keys():
        indices=collect_all_keys(self,r_triple)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: r_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: r_denominator(self,x),indices))))
        new_r=((self.prio_tau_r*self.prio_mu_r)+numerator_sum)/(self.prio_tau_r+denominator_sum)
        if self.debug: print(f"Old r {self.r[self.hash_r[r_triple]]}.", sys.stdout)
        self.r[self.hash_r[r_triple]]= new_r
        if self.debug: print(f"New r {self.r[self.hash_r[r_triple]]}.", sys.stdout)

def rr_loop(self):
    if self.verbose:
        print("Updating rr values",file=sys.stdout)
    for r_triple in self.hash_r.keys():
        indices=collect_all_keys(self,r_triple)
        numerator_sum=np.sum(np.asarray(list(map(lambda x: r_numerator(self,x),indices))))
        denominator_sum=np.sum(np.asarray(list(map(lambda x: r_denominator(self,x),indices))))
        new_rr=((self.prio_tau_rr*self.prio_mu_rr)+numerator_sum)/(self.prio_tau_rr+denominator_sum)
        if self.debug: print(f"Old rr {self.rr[self.hash_r[r_triple]]}.", sys.stdout)
        self.rr[self.hash_r[r_triple]]= new_rr
        if self.debug: print(f"New rr {self.rr[self.hash_r[r_triple]]}.", sys.stdout)

def r_numerator(self,key_dict):
    numerator = (self.tau[self.hash_t[key_dict["t"]]] * self.s[self.hash_s[key_dict["s"]]]) * \
                (self.D[self.hash_d[key_dict["t"]]] - \
                 (self.x[self.hash_x[key_dict["hj"]]] * self.y[self.hash_y[key_dict["hl"]]]) - \
                 (self.x[self.hash_x[key_dict["fk"]]] * self.y[self.hash_y[key_dict["fl"]]]) - \
                 (self.x[self.hash_x[key_dict["gi"]]] * self.y[self.hash_y[key_dict["gl"]]]))
    return numerator

def r_denominator(self,key_dict):
    denominator = self.tau[self.hash_t[key_dict["t"]]] * self.ss[self.hash_s[key_dict["s"]]]
    return denominator

def collect_all_keys(self,k_triple):
    #only get keys from tau and split them accordingly
    tau_keys = [x for x in self.hash_t.keys() if k_triple in x]
    out=[]
    for k in tau_keys:
        gi,hj,fk,l = k.split("_")
        gl = self.guide_to_gene(gi)
        hl = self.guide_to_gene(hj)
        fl = self.guide_to_gene(fk)
        s = find_s_keys(self,gl,hl,fl,l)
        r = find_r_keys(self,gi,hj,fk)
        out.append({"t":k,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s})
    return out


import numpy as np

def gene_to_guides(self,gene):
    sub = self.G3D.raw_anno_table.__deepcopy__()
    sub = sub.loc[(sub["gene_1"] == gene), :]
    sub = sub.loc[(sub["gene_2"] == gene), :]
    sub = sub.loc[(sub["gene_3"] == gene), :]
    return list(np.unique(sub[["guide_1", "guide_2", "guide_3"]]))

def gene_to_guides_fast(self,gene):
    if gene in self.gene_to_guide_dict:
        return self.gene_to_guide_dict[gene]
    else:
        return None

def guide_to_gene(self, guide):
    if guide in self.G3D.anno_dict_1:
        return self.G3D.anno_dict_1[guide]
    elif guide in self.G3D.anno_dict_2:
        return self.G3D.anno_dict_2[guide]
    elif guide in self.G3D.anno_dict_3:
        return self.G3D.anno_dict_3[guide]


def find_s_keys(self,gl,hl,fl,l): #for a given triple finds the s keys
    if "_".join([gl,hl,fl,l]) in self.hash_s:
        return "_".join([gl,hl,fl,l])
    else:
        for k in self.hash_s.keys():
            first,second,third,sample=k.split("_")
            if sample != l:
                continue
            elif gl == first and hl == second and fl == third:#123
                return k
            elif gl == second and hl == first and fl == third:#213
                return k
            elif gl == first and hl == third and fl == second:#132
                return k
            elif gl == second and hl == third and fl == first:#231
                return k
            elif gl == third and hl == first and fl == second:#312
                return k
            elif gl == third and hl == second and fl == first:#321
                return k
        return None

def find_r_keys(self,gi,hj,fk):#for a given triple finds the r keys
    if "_".join([gi,hj,fk]) in self.hash_r:
        return "_".join([gi,hj,fk])
    else:
        for k in self.hash_r.keys():
            first,second,third=k.split("_")
            if gi == first and hj == second and fk == third:#123
                return k
            elif gi == second and hj == first and fk == third:#213
                return k
            elif gi == first and hj == third and fk == second:#132
                return k
            elif gi == second and hj == third and fk == first:#231
                return k
            elif gi == third and hj == first and fk == second:#312
                return k
            elif gi == third and hj == second and fk == first:#321
                return k
        return None

def decreasing(self,L):
    '''
    Checks if list L is decreasing. NOT strictly decreasing.
    :param L:
    :return:
    '''
    return all(x>=y for x, y in zip(L, L[1:]))


import numpy as np
from Gemi3.pyCAVI.util import find_s_keys,find_r_keys

def calc_MAE(self):
    indices=collect_all_keys(self)
    MAE=np.mean(np.abs(np.asarray(list(map(lambda x: MAE_one_triple(self,x),indices)))))
    return MAE


def MAE_one_triple(self,key_dict):
    out=self.D[self.hash_d[key_dict["t"]]]- \
        self.x[self.hash_x[key_dict["gi"]]] * self.y[self.hash_y[key_dict["gl"]]]- \
        self.x[self.hash_x[key_dict["hj"]]] * self.y[self.hash_y[key_dict["hl"]]]- \
        self.x[self.hash_x[key_dict["fk"]]] * self.y[self.hash_y[key_dict["fl"]]]- \
        self.r[self.hash_r[key_dict["r"]]]*self.s[self.hash_s[key_dict["s"]]]
    return out

def collect_all_keys(self):
    #only get keys from tau and split them accordingly
    tau_keys = [x for x in self.hash_t.keys()]
    out=[]
    for k in tau_keys:
        gi,hj,fk,l = k.split("_")
        gl = self.guide_to_gene(gi)
        hl = self.guide_to_gene(hj)
        fl = self.guide_to_gene(fk)
        s = find_s_keys(self,gl,hl,fl,l)
        r = find_r_keys(self,gi,hj,fk)
        out.append({"t":k,"gi":gi,"hj":hj,"fk":fk,"l":l,
                    "gl":"_".join([gl,l]),
                    "hl":"_".join([hl,l]),
                    "fl":"_".join([hl,l]),
                    "r":r,
                    "s":s})
    return out


import pandas as pd
import numpy as np
import sys
import copy


def calc_observed_LFC(raw_data,ETP,replicates,normalize,constant=32):
    '''
    calculates the LFC score between early and late timepoint

    :param raw_data: df with raw count data, optionally normalized to million reads
    :param ETP: list with early timepoint column names
    :param replicates: replicate annotation
    :param normalize: bool, True or False
    :param constant: constant to add during LFC calculation, default 32
    :return: df with LFC table
    '''
    #raw_data="/Users/lukasmadenach/Desktop/Paper/Zhou2020/gemi3_on_zhou_data/count_table_2d_zhou.txt"
    #ETP=["Cell_D15.rep1","Cell_D15.rep2" ]
    #annotations="/Users/lukasmadenach/Desktop/Paper/Zhou2020/gemi3_on_zhou_data/annotation_2d_zhou.txt"
    #replicates="/Users/lukasmadenach/Desktop/Paper/Zhou2020/gemi3_on_zhou_data/replicates_2D_zhou.txt"
    raw_df=pd.read_csv(raw_data,sep="\t",index_col=0)
    rep_anno=pd.read_csv(replicates,sep="\t") #concatenated sample and rep must be separated by "."

    if normalize:
        total_counts=raw_df.sum(skipna=True).sum(skipna=True)
        scale=total_counts/raw_df.shape[1]
        raw_df=raw_df.apply(lambda x: ((x / x.sum())*scale)+constant, axis=0)

    raw_df=raw_df.apply(lambda x: np.log2(x)-np.log2(np.nanmedian(x)),axis=0)
    ETP_col=raw_df.loc[:,ETP].mean(axis=1)
    ETP_col.rename(ETP[0].split(".")[0],inplace=True)
    LFC_out=pd.DataFrame(ETP_col)
    rep_anno = rep_anno.loc[~rep_anno["colname"].isin(ETP), :]

    for s in set(rep_anno["samplename"]):
        mask=rep_anno.loc[rep_anno["samplename"]==s,:]["colname"]
        sample_col = raw_df.loc[:, list(mask)].mean(axis=1)
        sample_col.rename(s,inplace=True)
        LFC_out[s]=sample_col

    for s in LFC_out.columns:
        LFC_out[s]=LFC_out[s].subtract(ETP_col)
    LFC_out=LFC_out.drop(ETP[0].split(".")[0],axis=1)
    return LFC_out

class G3D():

    def __init__(self,LFC_table, annotation_table,samples,verbose=False):
        '''
        reads file with LFC table, also needs annotation table example files in Gemi3/data

        :param LFC_table: table wiht logfoldchange data, each column is one sample, index is guide RNA combinations, example in Gemi3/data
        :param annotation_table: table with guide annotation, used for matching guide RNA to genename, example in Gemi3/data
        :return: two values, LFC table values as pandas dataframe (df) and as tensorflow tf.connstant (observed)
        '''
        self.LFC_table=LFC_table
        self.samples=samples
        if "_" in samples:
            raise ValueError(" _ must not be in the sample names")
        self.annotation_table=annotation_table
        self.verbose=verbose
        if verbose: print("Loading LFC table from file.",sys.stdout)
        self.full_df = pd.read_csv(LFC_table, sep="\t",index_col=0)  # table with log fold changes between early and late timepoint
        if verbose: print("Loading annotation table from file.", sys.stdout)
        self.raw_anno_table,self.anno_dict_1, self.anno_dict_2, self.anno_dict_3 = self.read_annotation(annotation_table)
        if verbose: print("Merging annotation with LFC table from file.", sys.stdout)
        self.gene_pairs1 = [self.gRNApair_to_genenames(name, self.anno_dict_1, self.anno_dict_2, self.anno_dict_3) for name in self.full_df.index]
        self.full_df["first_gene"] = [x[0] for x in self.gene_pairs1]
        self.full_df["second_gene"] = [x[1] for x in self.gene_pairs1]
        self.full_df["third_gene"] = [x[2] for x in self.gene_pairs1]
        self.raw_data = self.full_df.loc[:, self.samples].values

    def __copy__(self):
        return G3D(self.LFC_table, self.annotation_table, self.samples,self.verbose)

    def __deepcopy__(self,memo):
        return G3D(copy.deepcopy(self.LFC_table,memo),copy.deepcopy(self.annotation_table,memo),
                   copy.deepcopy(self.samples,memo),copy.deepcopy(self.verbose,memo))

    def read_annotation(self,annotation_table):
        '''
        function for reading annotation table

        :param annotation_table: table with guide annotation, used for matching guide RNA to genename, example in Gemi3/data
        :return: three dicts with guide RNA as key and gene name as value, one dict for each spot in 3D vector
        '''
        annotation = pd.read_csv(annotation_table, sep="\t",index_col=0)
        if not all([x in annotation.columns for x in ["guide_1","guide_2","guide_3","gene_1","gene_2","gene_3"]]):
            print("There should be the columns guide_1,guide_2,guide_3,gene_1,gene_2 and gene_3",file=sys.stderr)
            raise ValueError
        _anno_dict_1 = {k: v for k, v in zip(annotation["guide_1"], annotation["gene_1"])}
        _anno_dict_2 = {k: v for k, v in zip(annotation["guide_2"], annotation["gene_2"])}
        _anno_dict_3 = {k: v for k, v in zip(annotation["guide_3"], annotation["gene_3"])}
        return annotation,_anno_dict_1, _anno_dict_2, _anno_dict_3

    def gRNApair_to_genenames(self,name, dict1, dict2, dict3):
        '''
        function to replace guide RNA with genenames

        :param name: concatenated guide RNAs, separated by ";", usually from index of dataframe
        :param dict1: dict with guide RNA as key and genenames as value
        :param dict2: dict with guide RNA as key and genenames as value
        :param dict3: dict with guide RNA as key and genenames as value
        :return: three values, input string split into separate guide RNA and guide RNA replaced by genename
        '''
        p1, p2, p3 = name.split(";")
        if p1 in dict1.keys() and p2 in dict2.keys() and p3 in dict3.keys():
            return dict1[p1], dict2[p2], dict3[p3]
        elif p2 in dict1.keys() and p1 in dict2.keys() and p3 in dict3.keys():
            return dict1[p2], dict2[p1], dict3[p3]
        elif p1 in dict1.keys() and p3 in dict2.keys() and p2 in dict3.keys():
            return dict1[p1], dict2[p3], dict3[p2]
        elif p3 in dict1.keys() and p2 in dict2.keys() and p1 in dict3.keys():
            return dict1[p3], dict2[p2], dict3[p1]


    def set_raw_df(self,new_data):
        '''
        function sets new_data in self.raw_data

        :param new_data: pandas dataframe with new values
        :return: -
        '''
        if isinstance(new_data, pd.DataFrame):
            if self.raw_data.shape == new_data.values.shape:
                self.raw_data = new_data.values
                self.set_full_df(self.raw_data)
            else:
                print("Dimensions don't match", sys.stderr)

    def set_full_df(self,new):
        '''
        function sets new in self.full_df together with columns first_gene, second_gene and third_gene

        :param new:  dataframe
        :return: -
        '''
        if self.full_df.loc[:,self.samples].values.shape == new.shape:
            self.full_df=pd.DataFrame(data=new,index=self.full_df.index,columns=self.samples)
            self.full_df["first_gene"] = [x[0] for x in self.gene_pairs1]
            self.full_df["second_gene"] = [x[1] for x in self.gene_pairs1]
            self.full_df["third_gene"] = [x[2] for x in self.gene_pairs1]
        else:
            print("Dimensions don't match",sys.stderr)


import numpy as np
import sys
from . import g3datahandling as g3d
import itertools as itt
import math
from Gemi3.rsCAVI import start_inference


class Model1:
    from Gemi3.pyCAVI._update_x import x_loop,xx_loop #modules with coordinate update equations
    from Gemi3.pyCAVI._update_y import y_loop,yy_loop
    from Gemi3.pyCAVI._update_r import r_loop,rr_loop
    from Gemi3.pyCAVI._update_s import s_loop,ss_loop
    from Gemi3.pyCAVI._update_tau import tau_loop
    from Gemi3.pyCAVI._update_MAE import calc_MAE
    from Gemi3.pyCAVI.util import gene_to_guides,decreasing,guide_to_gene,gene_to_guides_fast  #guide_to_gene and gene_to_guides_fast functions needed as method for other scripts to access

    def __init__(self,G3D,ncgenes,verbose=False,debug=False,force=False,MAX_ITERATIONS=30,MAE_THRESHOLD=0.001,
                 prio_mu_x=1.0,prio_tau_x=1.0,prio_mu_xx=1.0,prio_tau_xx=1.0,
                 prio_mu_y=0.0,prio_tau_y=10.0,prio_mu_yy=0.0,prio_tau_yy=10.0,
                 prio_mu_r=1.0,prio_tau_r=1.0,prio_mu_rr=1.0,prio_tau_rr=1.0,
                 prio_mu_s=0.0,prio_tau_s=10.0,prio_mu_ss=0.0,prio_tau_ss=10.0):
        if not isinstance(G3D,g3d.G3D): # check if data is in correct data handling object
            raise TypeError("G3D object needed as input")
        self.G3D = G3D #set field with full datahandling object
        self.debug=debug #set bool for debug mode
        if self.debug: #if debug mode verbose is automatically true
            print("!!!!MODEL initialized in DEBUG mode!!!!\n!!!!Very verbose output will be generated!!!!")
            self.verbose=True
        else:
            self.verbose = verbose
        self.force=force #set bool for force, if force==true CAVI continues even if MAE increases after 1 step
        self.full_df = G3D.full_df #set field with pandas dataframe holding LFC and gene annotation
        self.MAX_ITERATIONS=MAX_ITERATIONS #set field with maximum number of iterations for CAVI
        self.MAE_THRESHOLD=MAE_THRESHOLD #set field with threshold. if change in MAE smaller than threshold CAVI stops
        self.samples = G3D.samples #set field with list of sample names
        self.is_primed = False #set state
        self.is_inferred = False #set inference state
        self.D = G3D.raw_data #set numpy array with LFC values
        if isinstance(ncgenes,list): #check if nc-genes were supplied and are correct type
            if len(ncgenes)==0:
                raise TypeError("ncgenes must be supplied")
            self.ncgenes=ncgenes
        else:
            raise TypeError("Ncgenes must be supplied as list of strings.")
        self.prio_mu_x=prio_mu_x #x and xx prio
        self.prio_tau_x=prio_tau_x
        self.prio_mu_xx = prio_mu_xx
        self.prio_tau_xx = prio_tau_xx
        self.prio_mu_y=prio_mu_y # y and yy prio
        self.prio_tau_y=prio_tau_y
        self.prio_mu_yy = prio_mu_yy
        self.prio_tau_yy = prio_tau_yy
        self.prio_mu_r=prio_mu_r # r and rr prio
        self.prio_tau_r=prio_tau_r
        self.prio_mu_rr = prio_mu_rr
        self.prio_tau_rr = prio_tau_rr
        self.prio_mu_s=prio_mu_s # s and ss prio
        self.prio_tau_s=prio_tau_s
        self.prio_mu_ss = prio_mu_ss
        self.prio_tau_ss = prio_tau_ss
        self.ncguides = list(itt.chain(*[self.gene_to_guides(x) for x in self.ncgenes])) #convert supplied ncgenes to guides and set
        if len(self.ncguides)==0:
            raise TypeError("Specified ncgenes not present")
        ncgenes_mask = self.full_df.first_gene.isin(self.ncgenes) | self.full_df.second_gene.isin(
            self.ncgenes) | self.full_df.third_gene.isin(self.ncgenes) #bool mask for removing rows containg ncgenes from df
        self.ncgenes_df = self.full_df.loc[ncgenes_mask, :] #set pandas df with only ncgenes
        self.full_df = self.full_df.loc[~ncgenes_mask, :] #remove all rows containing at least 1 nc gene
        only_two_target_genes_mask = self.full_df.apply(lambda x: x["first_gene"]==x["second_gene"] or x["first_gene"]==x["third_gene"] or x["second_gene"]==x["third_gene"],axis=1) #set mask to filter rows were only two different genes are targeted
        if self.verbose:
            print("Removing rows where only two different genes are targeted",file=sys.stdout)
            print(f"Removing {np.sum(only_two_target_genes_mask)} rows.",file=sys.stdout)
        self.full_df = self.full_df.loc[~only_two_target_genes_mask,:]
        if self.full_df.empty:
            raise ValueError("All rows only contain two different genes.")
        self.hash_x=dict()#initialze hash dict empty, later used in prime() method
        self.hash_y=dict()
        self.hash_r=dict()
        self.hash_s=dict()
        self.hash_t=dict()
        self.hash_d=dict()
        self.MAE=[] #set MAE as empty list
        unique_guides_tmp = list(map(lambda x: x.split(";"), self.full_df.index))
        self.unique_guides = [x for x in set(itt.chain(*unique_guides_tmp)) if x not in self.ncguides] #set list with unique guides removing ncguides
        del unique_guides_tmp
        self.unique_genes = [x.replace(" ","") for x in np.unique(self.full_df.loc[:, ["first_gene", "second_gene", "third_gene"]].values) if x not in self.ncgenes] #set list with unique genes (not including ncgenes) and removing whitespace from gene names
        self.gene_to_guide_dict={k:self.gene_to_guides(k) for k in self.unique_genes}
        self.guide_to_gene_dict=dict() #set guide to gene dict
        self.guide_to_gene_dict.update(G3D.anno_dict_1)
        self.guide_to_gene_dict.update(G3D.anno_dict_2)
        self.guide_to_gene_dict.update(G3D.anno_dict_3)


    def prime(self):
        '''
        functions sets values from input data, build hash tables, initializes variables from control genes

        :return: -
        '''
        unique_guides_len = len(self.unique_guides)
        unique_genes_len = len(self.unique_genes)
        unique_threeway_combinations_guides = len(self.full_df.loc[:, ["first_gene", "second_gene", "third_gene"]].index)
        unique_threeway_combinations_genes = math.comb((unique_genes_len),(3))

        #D values
        self.D=self.G3D.raw_data
        self.build_hashes("D") #hashes for D same as for tau

        #x values
        self.x = np.random.normal(self.prio_mu_x,np.reciprocal(self.prio_tau_x),(unique_guides_len))
        self.build_hashes("x")  # build hashes for x
        self.init_x_from_ncgenes() #init x values form ncgenes
        self.xx = self.x**2

        #y values
        self.y = np.random.normal(self.prio_mu_y,np.reciprocal(self.prio_tau_y),(unique_genes_len,len(self.samples)))
        self.build_hashes("y")  # build hashes for y
        self.init_y_from_ncgenes() #init y values form ncgenes
        self.yy = self.y ** 2

        #r values
        self.r = np.random.normal(self.prio_mu_r,np.reciprocal(self.prio_tau_r),(unique_threeway_combinations_guides))
        self.build_hashes("r")  # build hashes for r
        self.rr = self.r**2

        #s values
        self.s = np.random.normal(self.prio_mu_s,np.reciprocal(self.prio_tau_s),(unique_threeway_combinations_genes,len(self.samples)))
        self.build_hashes("s")  # build hashes for s
        self.ss = self.s**2

        #tau values
        self.tau = np.random.gamma(0.5,1.0,(len(self.full_df.index),len(self.samples)))
        self.build_hashes("t")  # build hashes for tau
        self.alpha = np.full((len(self.full_df.index),len(self.samples)),0.5)
        self.beta = np.full((len(self.full_df.index),len(self.samples)),1.0) #TODO write init function for beta

        #MAE values
        self.MAE.append(self.calc_MAE())
        if self.verbose:
            print(f"Starting MAE at {self.MAE[-1]}")
        self.is_primed=True

    def build_hashes(self,variable):
        '''
        Builds and stores hashes needed for faster CAVI

        :param variable:
        :return:
        '''
        unique_threeway_combinations_guides = self.full_df.loc[:, ["first_gene", "second_gene", "third_gene"]].index
        unique_threeway_combinations_genes = math.comb(len(self.unique_genes), (3))

        if variable=="x": #initialization removed
            if self.verbose: print("Building hashtable for x and xx",file=sys.stdout)
            if len(self.unique_guides) != len(self.x):
                #print("Error initializing x and xx",file=sys.stderr)
                raise ValueError("Error initializing x and xx")
            for guide,idx in zip(self.unique_guides,range(len(self.x))):
                self.hash_x.update({guide:idx})
            if self.debug: print(self.hash_x.keys(),sys.stdout)

        elif variable == "y": #initialization removed
            if self.verbose: print("Building hashtable for y and yy",file=sys.stdout)
            if self.y.shape != (len(self.unique_genes),len(self.samples)):
                #print("Error initializing y and yy",file=sys.stderr)
                raise ValueError("Error initializing y and yy")
            num_rows, num_cols = self.y.shape
            for guide,idx1 in zip(self.unique_genes,range(num_rows)):
                for sample,idx2 in zip(self.samples,range(num_cols)):
                    self.hash_y.update({"_".join([guide,sample]):(idx1,idx2)})
            if self.debug: print(self.hash_y.keys(), sys.stdout)

        elif variable == "r":#all guide combinations in specified order, initialization genes are removed
            if self.verbose: print("Building hashtable for r and rr", file=sys.stdout)
            if len(self.r) != len(unique_threeway_combinations_guides):
                #print("Error initializing r and rr", file=sys.stderr)
                raise ValueError("Error initializing r and rr")
            sorted_threeway_guides=[x.replace(";","_") for x in unique_threeway_combinations_guides]
            for guide,idx in zip(sorted_threeway_guides,range(len(self.r))):
                self.hash_r.update({guide:idx})
            if self.debug: print(self.hash_r.keys(), sys.stdout)

        elif variable == "s": #should not care about order of genes in combination, initialization genes are removed
            if self.verbose: print("Building hashtable for s and ss",file=sys.stdout)
            if self.s.shape != (unique_threeway_combinations_genes,len(self.samples)):
                #print("Error initializing s and ss", file=sys.stderr)
                raise ValueError("Error initializing s and ss")
            sorted_threeway_genes = np.sort(list(self.unique_genes))
            sorted_threeway_genes = itt.combinations(sorted_threeway_genes,3)
            num_rows, num_cols = self.s.shape
            for gene,idx1 in zip(sorted_threeway_genes,range(num_rows)):
                for sample,idx2 in zip(self.samples,range(num_cols)):
                    self.hash_s.update({"_".join(["_".join(gene),sample]):(idx1,idx2)})
            if self.debug: print(self.hash_r.keys(), sys.stdout)

        elif variable == "t":
            if self.verbose: print("Building hashtable for tau",file=sys.stdout)
            if self.tau.shape != (len(self.full_df.index),len(self.samples)):
                #print("Error initializing tau", file=sys.stderr)
                raise ValueError("Error initializing tau")
            num_rows, num_cols = self.tau.shape
            sorted_threeway_guides = [x.replace(";","_") for x in self.full_df.index]
            for guide,idx1 in zip(sorted_threeway_guides,range(num_rows)):
                for sample,idx2 in zip(self.samples,range(num_cols)):
                    self.hash_t.update({"_".join([guide,sample]):(idx1,idx2)})
            if self.debug: print(self.hash_t.keys(), sys.stdout)

        elif variable == "D":
            if self.verbose: print("Building hashtable for D",file=sys.stdout)
            num_rows, num_cols = self.D.shape
            sorted_threeway_guides = [x.replace(";","_") for x in self.full_df.index]
            for guide,idx1 in zip(sorted_threeway_guides,range(num_rows)):
                for sample,idx2 in zip(self.samples,range(num_cols)):
                    self.hash_d.update({"_".join([guide,sample]):(idx1,idx2)})
            if self.debug: print(self.hash_t.keys(), sys.stdout)

        else:
            print("Build hashes failed, should not happen, check source code",file=sys.stderr)
            raise KeyError("Wrong key for building hashes")

    def start_inference_python(self): #inferenc method, each variable gets updated while others are constant
        '''
        functions runs CAVI algorithm implemented with python
        !!!ATTENTION!!! faults are in this implementation of CAVI, use rsCAVI

        :return:
        '''
        print("!!!ATTENTION!!!\n!!!DEPRECATED!!!\nuse rsCAVI instead of this function",file=sys.stderr)
        if not self.is_primed:
            raise TypeError("Model must be primed first.")
        counter=0
        while counter <= self.MAX_ITERATIONS:
            self.x_loop()
            self.xx_loop()
            self.y_loop()
            self.yy_loop()
            self.r_loop()
            self.rr_loop()
            self.s_loop()
            self.ss_loop()
            self.tau_loop()
            self.MAE.append(self.calc_MAE())
            counter +=1
            if self.verbose:
                print(f"Iteration {counter} complete.",file=sys.stdout)
                print(f"MAE at {self.MAE[-1]}.",file=sys.stdout)
            if not self.force:
                if self.MAE[-1]>self.MAE[-2]:
                    print("Breaking CAVI: MAE increased with step. \n Init model with force=True if you want to keep going.", sys.stdout)
                    break
            if np.abs(self.MAE[-1]-self.MAE[-2])<self.MAE_THRESHOLD:
                if self.verbose: print("Breaking CAVI: MAE_THRESHOLD reached.",sys.stdout)
                break
            if self.MAE[-1] < 0:
                print("Breaking CAVI: MAE turned negative.", sys.stdout)
                break

    def start_inference_rust(self):
        '''
        functions runs CAVI algorithm implemented with rust

        :return:
        '''
        if not self.is_primed:
            raise TypeError("Model must be primed first.")
        if self.verbose:
            print("Handoff to Rust.", sys.stdout)
        result_from_rust = start_inference(self.verbose, self.debug, self.force, self.MAX_ITERATIONS, self.MAE_THRESHOLD,
                                   self.ncgenes, self.ncguides, self.samples, self.MAE,
                                   self.unique_guides, self.unique_genes, self.guide_to_gene_dict,
                                   self.gene_to_guide_dict,
                                   self.prio_mu_x, self.prio_tau_x, self.prio_mu_xx, self.prio_tau_xx,
                                   self.prio_mu_y, self.prio_tau_y, self.prio_mu_yy, self.prio_tau_yy,
                                   self.prio_mu_r, self.prio_tau_r, self.prio_mu_rr, self.prio_tau_rr,
                                   self.prio_mu_s, self.prio_tau_s, self.prio_mu_ss, self.prio_tau_ss,
                                   self.hash_x, self.x.tolist(), self.xx.tolist(), self.x.shape[0],
                                   self.hash_y, self.y.flatten().tolist(), self.yy.flatten().tolist(),
                                   self.y.shape[0], self.y.shape[1],
                                   self.hash_r, self.r.tolist(), self.rr.tolist(), self.r.shape[0],
                                   self.hash_s, self.s.flatten().tolist(), self.ss.flatten().tolist(),
                                   self.s.shape[0], self.s.shape[1],
                                   self.hash_t, self.tau.flatten().tolist(), self.tau.shape[0],
                                   self.tau.shape[1],
                                   self.alpha.flatten().tolist(), self.beta.flatten().tolist(),
                                   self.full_df[self.samples].to_numpy().flatten().tolist())
        if self.verbose:
            print("Result returned from Rust. Checking...", sys.stdout)
        if not self.decreasing(result_from_rust[0]):
            print("MAE increased with step. \n Discarding results.", sys.stdout)
            return
        if self.verbose:
            print("Check completed. Setting values.", sys.stdout)
        nrow_y = self.y.shape[0]
        ncol_y = self.y.shape[1]
        nrow_s = self.s.shape[0]
        ncol_s = self.s.shape[1]
        nrow_tau = self.tau.shape[0]
        ncol_tau = self.tau.shape[1]

        self.MAE = result_from_rust[0]
        self.x = np.array(result_from_rust[1])
        self.xx = np.array(result_from_rust[2])
        self.y = np.array(result_from_rust[3]).reshape((nrow_y,ncol_y))
        self.yy = np.array(result_from_rust[4]).reshape((nrow_y,ncol_y))
        self.r = np.array(result_from_rust[5])
        self.rr = np.array(result_from_rust[6])
        self.s = np.array(result_from_rust[7]).reshape((nrow_s,ncol_s))
        self.ss = np.array(result_from_rust[8]).reshape((nrow_s,ncol_s))
        self.tau = np.array(result_from_rust[9]).reshape((nrow_tau,ncol_tau))
        self.alpha = np.array(result_from_rust[10]).reshape((nrow_tau,ncol_tau))
        self.beta = np.array(result_from_rust[11]).reshape((nrow_tau,ncol_tau))
        self.is_inferred = True #set new inference state


    def init_x_from_ncgenes(self):
        '''
        calc initial values for x variables from specified control genes

        :return:
        '''
        counter = 0
        for guide in self.hash_x.keys():
            counter += 1
            mask = [x for x in self.ncgenes_df.index if guide in x]
            subdf = self.ncgenes_df.loc[mask, :] #subdf with guide in every row at least once
            two_nc_genes_mask = subdf.apply(
                lambda x: x["first_gene"] in self.ncgenes and x["second_gene"] in self.ncgenes or x[
                    "first_gene"] in self.ncgenes and x["third_gene"] in self.ncgenes or x[
                              "second_gene"] in self.ncgenes and x["third_gene"] in self.ncgenes, axis=1)
            subdf = subdf.loc[two_nc_genes_mask, :]
            if subdf.empty:
                continue
            init_x_value = np.median(subdf.loc[:,self.samples].values)
            self.x[self.hash_x[guide]] = init_x_value
        if counter != len(self.hash_x.keys()):
            print("WARNING: Not all genes paired with 2 ncgenes",sys.stderr)


    def init_y_from_ncgenes(self):
        '''
        calc initial values for y variables from specified control genes

        :return:
        '''
        counter = 0
        for gene_sample in self.hash_y.keys():
            counter += 1
            gene,sample=gene_sample.split("_")
            mask = self.ncgenes_df.first_gene.isin([gene]) | self.ncgenes_df.second_gene.isin([gene]) | self.ncgenes_df.third_gene.isin([gene])
            subdf = self.ncgenes_df.loc[mask,:] #subdf with gene at least once in every row
            two_nc_genes_mask = subdf.apply(
                lambda x: x["first_gene"] in self.ncgenes and x["second_gene"] in self.ncgenes or x["first_gene"] in self.ncgenes and x["third_gene"] in self.ncgenes or x[
                    "second_gene"] in self.ncgenes and x["third_gene"] in self.ncgenes, axis=1)
            subdf =subdf.loc[two_nc_genes_mask,:]
            if subdf.empty:
                continue
            init_y_value=np.median(subdf.loc[:,sample].values)
            self.y[self.hash_y[gene_sample]]=init_y_value
        if counter != len(self.hash_y.keys()):
            print("WARNING: Not all genes paired with 2 ncgenes",sys.stderr)




from setuptools import setup
from setuptools_rust import Binding, RustExtension

setup(
    name="Gemi3",
    version="1.0",
    rust_extensions=[RustExtension("Gemi3.rsCAVI", binding=Binding.PyO3)],
    packages=["Gemi3","Gemi3.pyCAVI"],
    # rust extensions are not zip safe, just like C-extensions.
    zip_safe=False,
)